@@ -37,18 +37,11 @@ logging = climate.get_logger(__name__)
 FLOAT = theano.config.floatX
 
 
-def load(klass, filename, **kwargs):
+def load(filename, **kwargs):
     '''Load an entire network from a pickle file on disk.
 
     Parameters
     ----------
-    klass : callable
-        The constructor for a Network subclass (e.g., Autoencoder, Regressor,
-        etc.). This constructor will be invoked using keyword arguments loaded
-        from the given pickle file (and updated using any keyword arguments
-        passed to this load() invocation). The newly created network will then
-        load saved parameters (weights and biases) from the pickle file.
-
     filename : str
         Load the keyword arguments and parameters of a network from a pickle
         file at the named path. If this name ends in ".gz" then the input will
@@ -63,10 +56,11 @@ def load(klass, filename, **kwargs):
     '''
     opener = gzip.open if filename.lower().endswith('.gz') else open
     handle = opener(filename, 'rb')
-    kw = pickle.load(handle)['kwargs']
+    pkl = pickle.load(handle)
     handle.close()
+    kw = pkl['kwargs']
     kw.update(kwargs)
-    net = klass(**kw)
+    net = pkl['klass'](**kw)
     net.load_params(filename)
     return net
 
@@ -92,11 +86,11 @@ class Network(object):
         "hidden" layer with 20 units, and one "output" layer with 3 units. That
         is, inputs should be of length 10, and outputs will be of length 3.
 
-    hidden_activation : str
+    hidden_activation : str, optional
         The name of an activation function to use on hidden network units.
         Defaults to 'sigmoid'.
 
-    output_activation : str
+    output_activation : str, optional
         The name of an activation function to use on output units. Defaults to
         'linear'.
 
@@ -117,15 +111,15 @@ class Network(object):
     hidden_dropouts : float in [0, 1], optional
         Proportion of hidden unit activations to randomly set to 0.
 
-    decode : positive int, optional
+    decode_from : positive int, optional
         Any of the hidden layers can be tapped at the output. Just specify a
         value greater than 1 to tap the last N hidden layers. The default is 1,
         which decodes from just the last layer.
 
     tied_weights : bool, optional
         Construct decoding weights using the transpose of the encoding weights
-        on corresponding layers. If not True, decoding weights will be
-        constructed using a separate weight matrix.
+        on corresponding layers. Defaults to False, which means decoding weights
+        will be constructed using a separate weight matrix.
 
     Attributes
     ----------
@@ -142,26 +136,9 @@ class Network(object):
         Computed Theano variables representing the pre-activation inputs for
         network units.
 
-    layers : tuple of int
-        A list of the numbers of units in each of the layers of the network. The
-        zero element of this tuple is the size of the input, the final element
-        is the size of the output, and the intervening numbers are the sizes of
-        the hidden layers.
-
-    hidden_activation : str
-        A string describing the hidden unit activation function.
-
-    output_activation : str
-        A string describing the output unit activation function.
-
-    tied_weights : bool
-        True iff this is an autoencoder model with tied decoding weights.
-
-    decode_from : int
-        An integer describing the number of hidden layers from which to "decode"
-        the value of the output layer. By default this is 1, which corresponds
-        to a traditional feedforward neural network, in which the "topmost"
-        hidden layer is the only layer that feeds into the output.
+    kwargs : dict
+        A dictionary containing the keyword arguments used to construct the
+        network.
     '''
 
     def __init__(self, **kwargs):
@@ -170,15 +147,10 @@ class Network(object):
         self.weights = []
         self.biases = []
         self.updates = {}
-
+        self.kwargs = kwargs
         self.rng = kwargs.get('rng') or RandomStreams()
 
-        self.layers = tuple(kwargs['layers'])
-        self.tied_weights = bool(kwargs.get('tied_weights'))
-        self.decode_from = int(kwargs.get('decode', 1))
-
-        self.hidden_activation = kwargs.get('hidden_activation') or \
-                                 kwargs.get('activation', 'sigmoid')
+        self.hidden_activation = kwargs.get('hidden_activation', 'logistic')
         self._hidden_func = self._build_activation(self.hidden_activation)
         if hasattr(self._hidden_func, '__theanets_name__'):
             logging.info('hidden activation: %s', self._hidden_func.__theanets_name__)
@@ -189,105 +161,136 @@ class Network(object):
             logging.info('output activation: %s', self._output_func.__theanets_name__)
 
         self.setup_vars()
-        _, encode_count = self.setup_encoder(**kwargs)
-        self.y, decode_count = self.setup_decoder(**kwargs)
-
-        logging.info('%d total network parameters', encode_count + decode_count)
+        self.setup_layers(**kwargs)
 
     def setup_vars(self):
         '''Setup Theano variables for our network.'''
         # x is a proxy for our network's input, and y for its output.
         self.x = TT.matrix('x')
 
-    def setup_encoder(self, **kwargs):
-        '''Set up a computation graph to map the input to layer activations.
+    def setup_layers(self, **kwargs):
+        '''Set up a computation graph for our network.
 
         Parameters
         ----------
         input_noise : float, optional
             Standard deviation of desired noise to inject into input.
-
         hidden_noise : float, optional
             Standard deviation of desired noise to inject into hidden unit
             activation output.
-
         input_dropouts : float in [0, 1], optional
             Proportion of input units to randomly set to 0.
-
         hidden_dropouts : float in [0, 1], optional
             Proportion of hidden unit activations to randomly set to 0.
+        decode_from : int, optional
+            Compute the activation of the output vector using the activations of
+            the last N hidden layers in the network. Defaults to 1, which
+            results in a traditional setup that decodes only from the
+            penultimate layer in the network.
+        '''
+        count = 0
+
+        # add noise to inputs.
+        x = self._add_noise(self.x,
+                            kwargs.get('input_noise', 0),
+                            kwargs.get('input_dropouts', 0))
+
+        # setup "encoder" layers.
+        kw = dict(
+            noise=kwargs.get('hidden_noise', 0),
+            dropout=kwargs.get('hidden_dropouts', 0),
+        )
+        sizes = self.get_encoder_layers()
+        for i, (nin, nout) in enumerate(zip(sizes[:-1], sizes[1:])):
+            z = self.hiddens and self.hiddens[-1] or x
+            count += self.add_feedforward_layer(z, nin, nout, label=i, **kw)
+
+        count += self.setup_decoder(**kwargs)
+
+        logging.info('%d total network parameters', count)
+
+    def add_feedforward_layer(self, input, nin, nout, **kwargs):
+        '''Add a new feedforward layer to the network.
+
+        Parameters
+        ----------
+        input : theano variable
+            The theano variable that represents the inputs to this layer.
+        nin : int
+            The number of input units to this layer.
+        nout : out
+            The number of output units from this layer.
+        label : any, optional
+            The name of this layer, used for logging and as the theano variable
+            name suffix. Defaults to the index of this layer in the network.
+        noise : float, optional
+            Add zero-mean gaussian noise to the output activation of hidden
+            units. Defaults to 0 (no noise).
+        dropout : float, optional
+            Simulate "dropout" in this layer by setting the given fraction of
+            output activations randomly to zero. Defaults to 0 (no dropout).
 
         Returns
         -------
-        x : Theano variable
-            A variable representing the (possibly noisy) input vector.
-        parameter_count : int
-            The number of parameters created in the forward map.
+        count : int
+            A count of the number of learnable parameters in this layer.
         '''
-        sizes = self.check_layer_sizes()
-        parameter_count = 0
-        x = z = self._add_noise(
-            self.x,
-            kwargs.get('input_noise', 0.),
-            kwargs.get('input_dropouts', 0.))
-        for i, (a, b) in enumerate(zip(sizes[:-1], sizes[1:])):
-            W, b, count = self.create_layer(a, b, i)
-            parameter_count += count
-            self.preacts.append(TT.dot(z, W) + b)
-            self.hiddens.append(self._add_noise(
-                self._hidden_func(self.preacts[-1]),
-                kwargs.get('hidden_noise', 0.),
-                kwargs.get('hidden_dropouts', 0.)))
-            self.weights.append(W)
-            self.biases.append(b)
-            z = self.hiddens[-1]
-        return x, parameter_count
+        label = kwargs.get('label') or len(self.hiddens)
+        W, nw = self.create_weights(nin, nout, label)
+        b, nb = self.create_bias(nout, label)
+        pre = TT.dot(input, W) + b
+        out = self._add_noise(self._hidden_func(pre),
+                              kwargs.get('noise', 0),
+                              kwargs.get('dropout', 0))
+        self.weights.append(W)
+        self.biases.append(b)
+        self.preacts.append(pre)
+        self.hiddens.append(out)
+        return nw + nb
 
     def setup_decoder(self, **kwargs):
         '''Set up the "decoding" computations from layer activations to output.
+
+        Parameters
+        ----------
+        decode_from : int, optional
+            Compute the activation of the output vector using the activations of
+            the last N hidden layers in the network. Defaults to 1, which
+            results in a traditional setup that decodes only from the
+            penultimate layer in the network.
+
+        Returns
+        -------
+        count : int
+            The number of parameters created in the decoding map.
         '''
-        parameter_count = 0
+        count = 0
+        B = len(self.biases) - 1
+        nout = self.layers[-1]
+        decoders = []
+        for i in range(B, B - kwargs.get('decode_from', 1), -1):
+            nin = self.biases[i].get_value(borrow=True).shape[0]
+            W, n = self.create_weights(nin, nout, 'out_%d' % i)
+            count += n
+            decoders.append(TT.dot(self.hiddens[i], W))
+            self.weights.append(W)
+        bias = theano.shared(np.zeros((nout, ), FLOAT), name='bias_out')
+        count += nout
+        pre = sum(decoders) + bias
+        self.biases.append(bias)
+        self.preacts.append(pre)
+        self.y = self._output_func(pre)
+        return count
 
-        if self.tied_weights:
-            for i in range(len(self.weights) - 1, -1, -1):
-                h = self.hiddens[-1]
-                a, b = self.weights[i].get_value(borrow=True).shape
-                logging.info('tied weights from layer %d: %s x %s', i, b, a)
-                o = theano.shared(np.zeros((a, ), FLOAT), name='b_out{}'.format(i))
-                self.preacts.append(TT.dot(h, self.weights[i].T) + o)
-                func = self._output_func if i == 0 else self._hidden_func
-                self.hiddens.append(func(self.preacts[-1]))
+    def get_encoder_layers(self):
+        '''Compute the layers that will be part of the network encoder.
 
-        else:
-            B = len(self.biases) - 1
-            n = self.layers[-1]
-            decoders = []
-            for i in range(B, B - self.decode_from, -1):
-                b = self.biases[i].get_value(borrow=True).shape[0]
-                Di, _, count = self.create_layer(b, n, 'out_%d' % i)
-                parameter_count += count - n
-                decoders.append(TT.dot(self.hiddens[i], Di))
-                self.weights.append(Di)
-            parameter_count += n
-            bias = theano.shared(np.zeros((n, ), FLOAT), name='bias_out')
-            self.biases.append(bias)
-            self.preacts.append(sum(decoders) + bias)
-            self.hiddens.append(self._output_func(self.preacts[-1]))
-
-        return self.hiddens.pop(), parameter_count
-
-    def check_layer_sizes(self):
-        # ensure that --layers is compatible with --tied-weights.
-        sizes = self.layers[:-1]
-        if self.tied_weights:
-            error = 'with --tied-weights, --layers must be an odd-length palindrome'
-            assert len(self.layers) % 2 == 1, error
-            k = len(self.layers) // 2
-            encode = np.asarray(self.layers[:k])
-            decode = np.asarray(self.layers[k+1:])
-            assert (encode == decode[::-1]).all(), error
-            sizes = self.layers[:k+1]
-        return sizes
+        Returns
+        -------
+        layers : list of int
+            A list of integers specifying sizes of the encoder network layers.
+        '''
+        return self.layers[:-1]
 
     @property
     def inputs(self):
@@ -303,9 +306,19 @@ class Network(object):
             yield 'h{}<0.1'.format(i+1), 100 * (abs(h) < 0.1).mean()
             yield 'h{}<0.9'.format(i+1), 100 * (abs(h) < 0.9).mean()
 
+    @property
+    def layers(self):
+        '''A tuple containing the layer configuration for this network.'''
+        return self.kwargs['layers']
+
+    @property
+    def tied_weights(self):
+        '''A boolean indicating this network uses tied weights.'''
+        return self.kwargs.get('tied_weights', False)
+
     @staticmethod
-    def create_layer(a, b, suffix, sparse=None):
-        '''Create a layer of weights and bias values.
+    def create_weights(a, b, suffix, **kwargs):
+        '''Create a layer of randomly-initialized weights.
 
         Parameters
         ----------
@@ -317,32 +330,65 @@ class Network(object):
             of "output" units that the weight matrix connects.
         suffix : str
             A string suffix to use in the Theano name for the created variables.
-            This string will be appended to 'W_' (for the weights) and 'b_' (for
-            the biases) parameters that are created and returned.
-        sparse : float in (0, 1)
-            If given, ensure that the weight matrix for the layer has only this
-            proportion of nonzero entries.
+            This string will be appended to 'W_' to name the parameters that are
+            created and returned.
+        sparse : float in (0, 1), optional
+            If given, ensure that the given fraction of the weight matrix is
+            set to zero. Defaults to 0, meaning all weights are nonzero.
+        radius : float, optional
+            If given, rescale the initial weights to have this spectral radius.
+            No scaling is performed by default.
 
         Returns
         -------
         weight : Theano shared array
             A shared array containing Theano values representing the weights
             connecting each "input" unit to each "output" unit.
-        bias : Theano shared array
-            A shared array containing Theano values representing the bias
-            values on each of the "output" units.
         count : int
             The number of parameters that are included in the returned
             variables.
         '''
         arr = np.random.randn(a, b) / np.sqrt(a + b)
-        if sparse is not None:
-            arr *= np.random.binomial(n=1, p=sparse, size=(a, b))
+        sparse = kwargs.get('sparse')
+        if sparse and 0 < sparse < 1:
+            k = min(a, b)
+            mask = np.random.binomial(n=1, p=1 - sparse, size=(a, b)).astype(bool)
+            mask[:k, :k] |= np.random.permutation(np.eye(k).astype(bool))
+            arr *= mask
+        radius = kwargs.get('radius')
+        if radius:
+            # rescale weights to have the appropriate spectral radius.
+            u, s, vT = np.linalg.svd(arr)
+            arr = np.dot(np.dot(u, np.diag(radius * s / abs(s[0]))), vT)
         weight = theano.shared(arr.astype(FLOAT), name='W_{}'.format(suffix))
-        arr = 1e-3 * np.random.randn(b)
-        bias = theano.shared(arr.astype(FLOAT), name='b_{}'.format(suffix))
         logging.info('weights for layer %s: %s x %s', suffix, a, b)
-        return weight, bias, (a + 1) * b
+        return weight, a * b
+
+    @staticmethod
+    def create_bias(b, suffix):
+        '''Create a vector of bias values.
+
+        Parameters
+        ----------
+        b : int
+            Number of units of bias to create.
+        suffix : str
+            A string suffix to use in the Theano name for the created variables.
+            This string will be appended to 'b_' to name the parameters that are
+            created and returned.
+
+        Returns
+        -------
+        bias : Theano shared array
+            A shared array containing Theano values representing the bias for a
+            set of computation units.
+        count : int
+            The number of parameters that are included in the returned
+            variables.
+        '''
+        arr = 1e-6 * np.random.randn(b)
+        bias = theano.shared(arr.astype(FLOAT), name='b_{}'.format(suffix))
+        return bias, b
 
     def _add_noise(self, x, sigma, rho):
         '''Add noise and dropouts to elements of x as needed.
@@ -439,6 +485,44 @@ class Network(object):
             params.extend(self.biases)
         return params
 
+    def get_weights(self, layer, borrow=False):
+        '''Return the current weights for a given layer.
+
+        Parameters
+        ----------
+        layer : int
+            The layer of weights to return.
+        borrow : bool, optional
+            Whether to "borrow" the reference to the weights. If True, this
+            returns a view onto the current weight array; if False (default), it
+            returns a copy of the weight array.
+
+        Returns
+        -------
+        ndarray :
+            The weight values, as a numpy array.
+        '''
+        return self.weights[layer].get_value(borrow=borrow)
+
+    def get_biases(self, layer, borrow=False):
+        '''Return the current bias vector for a given layer.
+
+        Parameters
+        ----------
+        layer : int
+            The layer of bias values to return.
+        borrow : bool, optional
+            Whether to "borrow" the reference to the biases. If True, this
+            returns a view onto the current bias vector; if False (default), it
+            returns a copy of the biases.
+
+        Returns
+        -------
+        ndarray :
+            The bias values, as a numpy vector.
+        '''
+        return self.biases[layer].get_value(borrow=borrow)
+
     def feed_forward(self, x):
         '''Compute a forward pass of all activations from the given input.
 
@@ -488,13 +572,7 @@ class Network(object):
         pickle.dump(dict(
             weights=[p.get_value().copy() for p in self.weights],
             biases=[p.get_value().copy() for p in self.biases],
-            kwargs=dict(
-                layers=self.layers,
-                hidden_activation=self.hidden_activation,
-                output_activation=self.output_activation,
-                tied_weights=self.tied_weights,
-                decode_from=self.decode_from,
-            )), handle, -1)
+            klass=self.__class__, kwargs=self.kwargs), handle, -1)
         handle.close()
         logging.info('%s: saved model parameters', filename)
 
@@ -565,6 +643,79 @@ class Network(object):
 class Autoencoder(Network):
     '''An autoencoder attempts to reproduce its input.'''
 
+    def setup_decoder(self, **kwargs):
+        '''Set up weights for the decoder layers of an autoencoder.
+
+        This implementation allows for weights to be tied to encoder weights.
+
+        Parameters
+        ----------
+        input_noise : float, optional
+            Standard deviation of desired noise to inject into input.
+        hidden_noise : float, optional
+            Standard deviation of desired noise to inject into hidden unit
+            activation output.
+        input_dropouts : float in [0, 1], optional
+            Proportion of input units to randomly set to 0.
+        hidden_dropouts : float in [0, 1], optional
+            Proportion of hidden unit activations to randomly set to 0.
+        tied_weights : bool, optional
+            If True, use decoding weights that are "tied" to the encoding
+            weights. This only makes sense for a limited set of "autoencoder"
+            layer configurations. Defaults to False.
+        decode_from : int, optional
+            For networks without tied weights, compute the activation of the
+            output vector using the activations of the last N hidden layers in
+            the network. Defaults to 1, which results in a traditional setup
+            that decodes only from the penultimate layer in the network.
+
+        Returns
+        -------
+        count : int
+            A count of the number of tunable decoder parameters.
+        '''
+        if not self.tied_weights:
+            return super(Autoencoder, self).setup_decoder(**kwargs)
+        count = 0
+        noise = kwargs.get('hidden_noise', 0)
+        dropout = kwargs.get('hidden_dropouts', 0)
+        for i in range(len(self.weights) - 1, -1, -1):
+            nin, nout = self.weights[i].get_value(borrow=True).shape
+            logging.info('tied weights from layer %d: %s x %s', i, nout, nin)
+            bias = theano.shared(np.zeros((nin, ), FLOAT), name='b_out{}'.format(i))
+            count += nin
+            pre = TT.dot(self.hiddens[-1], self.weights[i].T) + bias
+            act = self._add_noise(self._hidden_func(pre), noise, dropout)
+            if i == 0:
+                act = self._output_func(pre)
+            self.biases.append(bias)
+            self.preacts.append(pre)
+            self.hiddens.append(act)
+        self.y = self.hiddens.pop()
+        return count
+
+    def get_encoder_layers(self):
+        '''Compute the layers that will be part of the network encoder.
+
+        This implementation ensures that --layers is compatible with
+        --tied-weights.
+
+        Returns
+        -------
+        layers : list of int
+            A list of integers specifying sizes of the encoder network layers.
+        '''
+        sizes = self.layers[:-1]
+        if self.tied_weights:
+            error = 'with --tied-weights, --layers must be an odd-length palindrome'
+            assert len(self.layers) % 2 == 1, error
+            k = len(self.layers) // 2
+            encode = np.asarray(self.layers[:k])
+            decode = np.asarray(self.layers[k+1:])
+            assert (encode == decode[::-1]).all(), error
+            sizes = self.layers[:k+1]
+        return sizes
+
     @property
     def cost(self):
         err = self.y - self.x
