@@ -31,9 +31,7 @@ climate.add_arg('--help-optimize', action='store_true',
 g = climate.add_arg_group('Architecture')
 g.add_argument('-n', '--layers', nargs='+', type=int, metavar='N',
                help='construct a network with layers of size N1, N2, ...')
-g.add_argument('-g', '--activation', default='logistic', metavar='FUNC',
-               help='function for hidden unit activations DEPRECATED')
-g.add_argument('--hidden-activation', default='logistic', metavar='FUNC',
+g.add_argument('-g', '--hidden-activation', default='logistic', metavar='FUNC',
                help='function for hidden unit activations')
 g.add_argument('--output-activation', default='linear', metavar='FUNC',
                help='function for output unit activations')
@@ -47,8 +45,10 @@ g.add_argument('-O', '--optimize', default=(), nargs='+', metavar='ALGO',
                help='train with the given optimization algorithm(s)')
 g.add_argument('--no-learn-biases', action='store_true',
                help='if set, do not update bias parameters during learning')
+g.add_argument('--num-updates', type=int, default=10000, metavar='N',
+               help='perform at most N HF/scipy parameter updates')
 g.add_argument('-p', '--patience', type=int, default=50, metavar='N',
-               help='stop training if no improvement for N updates')
+               help='stop SGD/HF training if no improvement for N updates')
 g.add_argument('-v', '--validate', type=int, default=10, metavar='N',
                help='validate the model every N updates')
 g.add_argument('-b', '--batch-size', type=int, default=64, metavar='N',
@@ -59,6 +59,8 @@ g.add_argument('-V', '--valid-batches', type=int, metavar='N',
                help='use at most N batches during validation')
 g.add_argument('--save-progress', metavar='FILE',
                help='save the model periodically to FILE')
+g.add_argument('--save-every', type=float, default=0, metavar='N',
+               help='save the model every N iterations or -N minutes')
 
 g = climate.add_arg_group('Regularization')
 g.add_argument('--contractive-l2', type=float, default=0, metavar='S',
@@ -80,7 +82,7 @@ g.add_argument('--weight-l1', type=float, default=0, metavar='K',
 g.add_argument('--weight-l2', type=float, default=0, metavar='K',
                help='regularize network weights with K on the L2 term')
 
-g = climate.add_arg_group('SGD/NAG/RmsProp Optimization')
+g = climate.add_arg_group('SGD-Based Optimization')
 g.add_argument('-l', '--learning-rate', type=float, default=1e-4, metavar='V',
                help='train the network with a learning rate of V')
 g.add_argument('-m', '--momentum', type=float, default=0.9, metavar='V',
@@ -88,6 +90,12 @@ g.add_argument('-m', '--momentum', type=float, default=0.9, metavar='V',
 g.add_argument('--min-improvement', type=float, default=0.01, metavar='R',
                help='train until relative improvement is less than R')
 
+g = climate.add_arg_group('RmsProp Optimization')
+g.add_argument('--rms-clip', type=float, default=1000, metavar='V',
+               help='clip gradient values outside [-V, V]')
+g.add_argument('--rms-ema', type=float, default=0.9, metavar='V',
+               help='use an exponential moving average with weight V')
+
 g = climate.add_arg_group('Rprop Optimization')
 g.add_argument('--rprop-increase', type=float, default=1.01, metavar='R',
                help='increase parameter steps at rate R')
@@ -105,11 +113,13 @@ g.add_argument('--initial-lambda', type=float, default=1., metavar='K',
                help='start the HF method with Tikhonov damping of K')
 g.add_argument('--global-backtracking', action='store_true',
                help='backtrack to lowest cost parameters during CG')
-g.add_argument('--num-updates', type=int, default=10000, metavar='N',
-               help='perform at most N parameter updates')
 g.add_argument('--preconditioner', action='store_true',
                help='precondition the system during CG')
 
 g = climate.add_arg_group('Recurrent Nets')
 g.add_argument('--recurrent-error-start', type=int, default=3, metavar='T',
                help='compute network error starting at time T')
+g.add_argument('--recurrent-radius', type=float, default=1.1, metavar='R',
+               help='create recurrent weights with spectral radius R')
+g.add_argument('--recurrent-sparsity', type=float, default=0, metavar='R',
+               help='create recurrent weights with fraction R as 0')
