@@ -211,10 +211,11 @@ class NAG(SGD):
     technique, in the absence of a new gradient.
 
     In theory, this helps correct for oversteps during learning: If momentum
-    would lead us to overshoot, then the gradient at that overshot place
-    will point backwards, toward where we came from. (For details see
-    Sutskever, Martens, Dahl, and Hinton, ICML 2013, "On the importance of
-    initialization and momentum in deep learning.")
+    would lead us to overshoot, then the gradient at that overshot place will
+    point backwards, toward where we came from. (For details see Sutskever,
+    Martens, Dahl, and Hinton, ICML 2013, "On the importance of initialization
+    and momentum in deep learning." A PDF of this paper is freely available at
+    http://jmlr.csail.mit.edu/proceedings/papers/v28/sutskever13.pdf)
     '''
 
     def __init__(self, network, **kwargs):
@@ -342,16 +343,31 @@ class RmsProp(SGD):
     argument. If this weight is set to a low value, the EMA will have a short
     memory and will be prone to changing quickly. If the momentum parameter is
     set close to 1, the EMA will have a long history and will change slowly.
+
+    The implementation here is modeled after Graves (2013), "Generating
+    Sequences With Recurrent Neural Networks," http://arxiv.org/abs/1308.0850.
     '''
 
+    def __init__(self, network, **kwargs):
+        self.clip = kwargs.get('rms_clip', 1000)
+        self.ema = kwargs.get('rms_ema', 0.9)
+        super(RmsProp, self).__init__(network, **kwargs)
+
     def learning_updates(self):
         for param in self.params:
-            grad = TT.grad(self.J, param)
-            rms_ = theano.shared(
-                np.zeros_like(param.get_value()), name=param.name + '_rms')
-            rms = self.momentum * rms_ + (1 - self.momentum) * grad * grad
-            yield rms_, rms
-            yield param, param - self.learning_rate * grad / TT.sqrt(rms + 1e-8)
+            grad = TT.grad(self.J, param).clip(-self.clip, self.clip)
+            z = lambda: np.zeros_like(param.get_value())
+            g1_ = theano.shared(z(), name=param.name + '_g1')
+            g2_ = theano.shared(z(), name=param.name + '_g2')
+            vel_ = theano.shared(z(), name=param.name + '_vel')
+            g1 = self.ema * g1_ + (1 - self.ema) * grad
+            g2 = self.ema * g2_ + (1 - self.ema) * grad * grad
+            rms = TT.sqrt(g2 - g1 * g1 + 1e-4)
+            vel = self.momentum * vel_ - self.learning_rate * grad / rms
+            yield g1_, g1
+            yield g2_, g2
+            yield vel_, vel
+            yield param, param + vel
 
 
 class Scipy(Trainer):
@@ -363,6 +379,7 @@ class Scipy(Trainer):
         super(Scipy, self).__init__(network, **kwargs)
 
         self.method = method
+        self.iterations = kwargs.get('num_updates', 100)
 
         logging.info('compiling gradient function')
         self.f_grad = theano.function(network.inputs, TT.grad(self.J, self.params))
@@ -385,8 +402,7 @@ class Scipy(Trainer):
             costs = np.mean([self.f_eval(*x) for x in train_set], axis=0)
             cost_desc = ' '.join(
                 '%s=%.2f' % el for el in zip(self.cost_names, costs))
-            logging.info('scipy %s %i/%i %s',
-                         self.method, i + 1, self.iterations, cost_desc)
+            logging.info('scipy.%s %i %s', self.method, i + 1, cost_desc)
 
         for i in range(self.iterations):
             try:
@@ -594,7 +610,8 @@ class Layerwise(Trainer):
                     net.hiddens.append(TT.dot(net.hiddens[-1], weights[j].T))
                 net.y = net._output_func(net.hiddens.pop())
             else:
-                W, b, _ = net.create_layer(nhids[i-1], nout, 'layerwise')
+                W, _ = net.create_weights(nhids[i-1], nout, 'layerwise')
+                b, _ = net.create_bias(nout, 'layerwise')
                 net.weights = [weights[i-1], W]
                 net.biases = [biases[i-1], b]
                 net.y = net._output_func(TT.dot(hiddens[i-1], W) + b)
@@ -607,3 +624,46 @@ class Layerwise(Trainer):
         net.hiddens = hiddens
         net.weights = weights
         net.biases = biases
+
+
+class UnsupervisedPretrainer(Trainer):
+    '''Train a discriminative model using an unsupervised pre-training step.
+
+    This trainer is a bit of glue code that creates a "shadow" autoencoder based
+    on a current network model, trains the autoencoder, and then transfers the
+    trained weights back to the original model.
+
+    This code is intended mostly as a proof-of-concept; more elaborate training
+    strategies are certainly possible but should be coded outside the core
+    package.
+    '''
+
+    def __init__(self, network, *args, **kwargs):
+        self.network = network
+        self.args = args
+        self.kwargs = kwargs
+
+    def train(self, train_set, valid_set=None, **kwargs):
+        # construct a copy of the input network, with tied weights in an
+        # autoencoder configuration.
+        layers = self.network.layers[:-1]
+        ae = feedforward.Autoencoder(
+            tied_weights=True,
+            layers=layers[:-1] + layers[::-1],
+            hidden_activation=self.network.hidden_activation,
+            output_activation='linear')
+
+        # copy the current weights into the autoencoder.
+        for i in range(len(layers) - 1):
+            ae.weights[i].set_value(self.network.get_weights(i))
+            ae.biases[i].set_value(self.network.get_biases(i))
+
+        # train the autoencoder using a layerwise strategy.
+        pre = Layerwise(ae, *self.args, **self.kwargs)
+        for costs in pre.train(train_set, valid_set=valid_set, **kwargs):
+            yield costs
+
+        # copy the trained autoencoder weights into our original model.
+        for i in range(len(layers) - 1):
+            self.network.weights[i].set_value(ae.get_weights(i))
+            self.network.biases[i].set_value(ae.get_biases(i))
